/*
 * Copyright (C) 2015 by Multi-Tech Systems
 *
 * This file is part of libmts.
 *
 * libmts is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * libmts is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libmts.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*! \file   MTS_Condition.h
 \brief  A condition that can be waited on or signaled
 \date   15JUN11
 \author Sean Godinez

 A platform independent condition that can be waited on or signaled.
 */

#ifndef _MTS_CONDITION_H_
#define _MTS_CONDITION_H_

#include <mts/MTS_Stdint.h>
#include <mts/MTS_AutoPtr.h>
#include <mts/MTS_NonCopyable.h>

#ifdef WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif

namespace MTS {

    class Lock;

    //! A condition
    /*!
     A platform independent condition that can be waited on or signaled.  Useful in building semaphores.

     \sa Lock
     */
    class Condition: private NonCopyable {
            friend class Lock;

        public:
            ~Condition();                           //!< Destructs the Condition

            void wait();                            //!< Waits on *this Condition for signal
            void wait(uint32_t millis);             //!< Waits on *this Condition for signal or given amount of time
            void signal();                          //!< Signals *this Condition

        private:

            Lock* m_pLock;                          //!< Lock that created *this Condition
#ifdef WIN32
            AutoPtr<HANDLE> m_apCondition;          //!< Handle to WIN32 Condition
#else
            AutoPtr<pthread_cond_t> m_apCondition;  //!< Handle to PTHREAD Condition
#endif

            explicit Condition(Lock* lock);         //!< Explicit Constructor! Must pass Lock as parent.
    };
}

#endif
