/*
 * Copyright (C) 2015 by Multi-Tech Systems
 *
 * This file is part of libmts.
 *
 * libmts is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * libmts is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libmts.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*! \file   MTS_TimerThread.h
 \brief  An abstract worker thread
 \date   10MAY13
 \author Mike Fiore

 A platform independent thread that executes a task every x miliseconds.
 Thread can be paused and unpaused.
 */

#ifndef _MTS_TIMER_THREAD_H_
#define _MTS_TIMER_THREAD_H_

#include <mts/MTS_Thread.h>
#include <mts/MTS_AutoPtr.h>
#include <mts/MTS_Lock.h>

namespace MTS {

    //! An abstract worker thread
    /*!
     A platform independent worker thread class
     */

	class TimerThread: public Thread {

		public:

			~TimerThread();							//!< Destructs the thread

			virtual bool isExecuting() const;				//!< Returns true if the thread has been started

			virtual void stopExecution();				//!< Stops running execute() function

			virtual void restartExecution();			//!< Resumes running execute() function

			virtual void updateIntervalMillis(uint32_t newMillis);			//!< Updates interval to given value

		protected:

			explicit TimerThread(const std::string& name);									//!< Creates the thread

			explicit TimerThread(const std::string& name, uint32_t intervalMillis);			//!< Creates the thread

		private:

			void run();								//!< Abstracts the running of the thread
			virtual void execute() = 0;				//!< main() of the thread which is executed once per interval
			void init();							//!< Startup code common to all constructors

			AutoPtr<Lock> m_apStateLock;			//!< Condition Lock
			bool m_bCurrentlyExecuting;				//!< Currently executing flag
			bool m_bExecutionStopped;				//!< Execution stopped flag
			uint32_t m_ui32IntervalMillis;			//!< How often the thread executes
	};
}

#endif
